Imports UvARescue.Agent
Imports UvARescue.Tools

Imports System.IO
Imports System.Text
Imports System.Windows.Forms

Public Class TeamConfigDialog
    Inherits ConfigDialog

    Public Sub New(ByVal config As TeamConfig)
        MyBase.New(config)

        ' This call is required by the Windows Form Designer.
        InitializeComponent()

    End Sub

    Public ReadOnly Property TeamConfig() As TeamConfig
        Get
            Return DirectCast(MyBase.Config, TeamConfig)
        End Get
    End Property

    Protected Overrides Sub OnOpened(ByVal filename As String)
        MyBase.OnOpened(filename)
        My.Settings.LastTeamConfigFile = filename
        My.Settings.Save()
    End Sub
    Protected Overrides Sub OnSaved(ByVal filename As String)
        MyBase.OnSaved(filename)
        My.Settings.LastTeamConfigFile = filename
        My.Settings.Save()
    End Sub

    Public Property OperatorName() As String
        Get
            Return Me.txtOperatorName.Text
        End Get
        Set(ByVal value As String)
            Me.txtOperatorName.Text = value
        End Set
    End Property

    Public Property TeamMembers() As String
        Get
            Return Me.txtTeamMembers.Text
        End Get
        Set(ByVal value As String)
            Me.txtTeamMembers.Text = value
        End Set
    End Property

    Protected Overrides Sub SyncGuiWithConfig()
        MyBase.SyncGuiWithConfig()
        With Me.TeamConfig

            Me.txtLevelName.Text = .LevelName
            Me.txtStartPoses.Text = .StartPoses

            Me.txtLocalhost.Text = .LocalHost
            Me.txtUsarSimHost.Text = .UsarSimHost
            Me.txtUsarSimPort.Text = CStr(.UsarSimPort)
            Me.txtImageServerHost.Text = .ImageServerHost
            Me.txtImageServerPort.Text = CStr(.ImageServerPort)
            Me.txtWirelessServerHost.Text = .WirelessServerHost
            Me.txtWirelessServerPort.Text = CStr(.WirelessServerPort)
            Me.chkboxBackup.Checked = .CreateBackups
            Me.txtBackupFrequency.Text = CStr(.BackupFrequency)
        End With
    End Sub

    Protected Overrides Sub SyncConfigWithGui()
        MyBase.SyncConfigWithGui()
        With Me.TeamConfig

            .LevelName = Me.txtLevelName.Text
            .StartPoses = Me.txtStartPoses.Text

            .OperatorName = Me.txtOperatorName.Text
            .TeamMembers = Me.txtTeamMembers.Text

            .LocalHost = Me.txtLocalhost.Text

            .UsarSimHost = Me.txtUsarSimHost.Text
            If IsNumeric(Me.txtUsarSimPort.Text) Then
                .UsarSimPort = CInt(Double.Parse(Me.txtUsarSimPort.Text))
            End If

            .ImageServerHost = Me.txtImageServerHost.Text
            If IsNumeric(Me.txtImageServerPort.Text) Then
                .ImageServerPort = CInt(Double.Parse(Me.txtImageServerPort.Text))
            End If

            .WirelessServerHost = Me.txtWirelessServerHost.Text
            If IsNumeric(Me.txtWirelessServerPort.Text) Then
                .WirelessServerPort = CInt(Double.Parse(Me.txtWirelessServerPort.Text))
            End If

            .CreateBackups = Me.chkboxBackup.Checked
            If (IsNumeric(Me.txtBackupFrequency.Text)) Then
                .BackupFrequency = CInt(Double.Parse(Me.txtBackupFrequency.Text))
            End If

        End With
        Me.TeamConfig.LevelName = GetLevelName(Me.txtUsarSimHost.Text, Integer.Parse(Me.txtUsarSimPort.Text))
        Me.TeamConfig.StartPoses = GetStartPoses(Me.txtUsarSimHost.Text, Integer.Parse(Me.txtUsarSimPort.Text))
        Me.SyncGuiWithConfig()
    End Sub


    Public Shared Function GetStartPoses(ByVal host As String, ByVal port As Integer) As String

        Try
            Dim returnValue As String = Nothing
            Dim nmbStartPoses As Integer = -1
            Dim usarSimConnection As New TcpMessagingConnection


            

            usarSimConnection.Connect(host, port)
            'wait for the DeadMatch answer.

            Threading.Thread.Sleep(TimeSpan.FromSeconds(1))

            If Not usarSimConnection.IsConnected Then
                Console.WriteLine("Couldn't query UsarSim server at host '" + host + "' for StartPoses.")
                Return Nothing
            End If


            Dim msgs As Specialized.StringCollection = usarSimConnection.ReceiveMessages(4096)
            For Each msg As String In msgs
                'NFO {StartPoses 3} {BlueGoal -2.55,-0.00,-0.19 0.00,-0.03,-0.01 MiddleField 0.01,-0.00,-0.19 0.00,0.00,0.00 YellowGoal 2.57,-0.01,-0.19 0.00,0.00,3.13}
                If String.IsNullOrEmpty(msg) Then
                    Continue For
                End If

                Dim msgtype As MessageType = MessageType.Unknown
                Dim parts As New Specialized.StringCollection

                'extract msgtype first
                Dim i As Integer = 0
                Dim j As Integer = msg.IndexOf(" ")
                Dim part As String = Nothing
                If j > 0 Then
                    part = msg.Substring(i, j - i)
                End If
                If part.ToUpper = "NFO" Then
                    msgtype = MessageType.Info
                Else
                    Continue For
                End If

                i = msg.IndexOf("{", j)
                While i >= 0
                    j = msg.IndexOf("}", i)
                    If j >= 0 Then
                        part = msg.Substring(i + 1, j - i - 1)
                        parts.Add(part)

                        i = msg.IndexOf("{", j)

                    Else
                        'done
                        Exit While

                    End If
                    If part.StartsWith("Gametype") Then
                        Continue For
                    End If
                    If part.StartsWith("StartPoses") Then
                        nmbStartPoses = Integer.Parse(part.Substring(10)) 'nmbStartPoses not saved
                    Else
                        returnValue = part
                    End If
                    'Now SyncGuiWithConfig in AgentConfig
                End While

                
            Next

            ' Send the request

            Dim cmd As New StringBuilder()
            cmd.Append("GETSTARTPOSES")
            cmd.Append(Environment.NewLine) 'newline! very important!

            'simply use the standard interface for sending commands.
            usarSimConnection.Send(cmd.ToString)

            Threading.Thread.Sleep(TimeSpan.FromSeconds(1))

            msgs = usarSimConnection.ReceiveMessages(4096)
            For Each msg As String In msgs
                'NFO {StartPoses 3} {BlueGoal -2.55,-0.00,-0.19 0.00,-0.03,-0.01 MiddleField 0.01,-0.00,-0.19 0.00,0.00,0.00 YellowGoal 2.57,-0.01,-0.19 0.00,0.00,3.13}
                If String.IsNullOrEmpty(msg) Then
                    Continue For
                End If

                Dim msgtype As MessageType = MessageType.Unknown
                Dim parts As New Specialized.StringCollection

                'extract msgtype first
                Dim i As Integer = 0
                Dim j As Integer = msg.IndexOf(" ")
                Dim part As String = Nothing
                If j > 0 Then
                    part = msg.Substring(i, j - i)
                End If
                If part.ToUpper = "NFO" Then
                    msgtype = MessageType.Info
                Else
                    Continue For
                End If

                i = msg.IndexOf("{", j)
                While i >= 0
                    j = msg.IndexOf("}", i)
                    If j >= 0 Then
                        part = msg.Substring(i + 1, j - i - 1)
                        parts.Add(part)

                        i = msg.IndexOf("{", j)

                    Else
                        'done
                        Exit While

                    End If

                    If part.StartsWith("Gametype") Then
                        Continue For
                    End If

                    'For DM-Mapping_250: "PlayerStart 0.75,-7.28,-0.44 0.00,0.00,0.00 robot1 3.63,10.80,-0.18 0.00,0.00,-1.59 robot2 1.91,10.00,-0.18 0.00,0.00,0.00 Robot3 3.85,9.17,-0.18 0.00,0.00,1.66"	String

                    If part.StartsWith("StartPoses") Then
                        nmbStartPoses = Integer.Parse(part.Substring(10)) 'nmbStartPoses not saved
                    Else
                        returnValue = part
                    End If
                    'Now SyncGuiWithConfig in AgentConfig


                End While


            Next

            
            'AVI: search the parts for key StartPoses, and loop through all Poses

            'AVI: Make enumeration, which can be used as drop down for each robot.

            usarSimConnection.Disconnect()
            Return returnValue

        Catch ex As Exception
            MsgBox("Error occurred during GetStartPoses" & vbNewLine & ex.Message, MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
            Return Nothing
        End Try
    End Function

    Public Shared Function GetLevelName(ByVal host As String, ByVal port As Integer) As String

        Try
            Dim returnValue As String = Nothing
            Dim usarSimConnection As New TcpMessagingConnection

            usarSimConnection.Connect(host, port)

            Threading.Thread.Sleep(TimeSpan.FromSeconds(1))

            If Not usarSimConnection.IsConnected Then
                Console.WriteLine("Couldn't query UsarSim server at host '" + host + "' for Level name.")
                Return Nothing
            End If


            Dim msgs As Specialized.StringCollection = usarSimConnection.ReceiveMessages(4096)
            For Each msg As String In msgs
                'NFO {StartPoses 3} {BlueGoal -2.55,-0.00,-0.19 0.00,-0.03,-0.01 MiddleField 0.01,-0.00,-0.19 0.00,0.00,0.00 YellowGoal 2.57,-0.01,-0.19 0.00,0.00,3.13}
                If String.IsNullOrEmpty(msg) Then
                    Continue For
                End If

                Dim msgtype As MessageType = MessageType.Unknown
                Dim parts As New Specialized.StringCollection

                'extract msgtype first
                Dim i As Integer = 0
                Dim j As Integer = msg.IndexOf(" ")
                Dim part As String = Nothing
                If j > 0 Then
                    part = msg.Substring(i, j - i)
                End If
                If part.ToUpper = "NFO" Then
                    msgtype = MessageType.Info
                Else
                    Continue For
                End If

                i = msg.IndexOf("{", j)
                While i >= 0
                    j = msg.IndexOf("}", i)
                    If j >= 0 Then
                        part = msg.Substring(i + 1, j - i - 1)
                        parts.Add(part)

                        i = msg.IndexOf("{", j)

                    Else
                        'done
                        Exit While

                    End If

                    If part.StartsWith("Level") Then
                        returnValue = part.Substring(6)
                    End If
                    'Now SyncGuiWithConfig in AgentConfig
                End While


            Next

            'AVI: search the parts for key StartPoses, and loop through all Poses

            'AVI: Make enumeration, which can be used as drop down for each robot.

            usarSimConnection.Disconnect()
            Return returnValue

        Catch ex As Exception
            MsgBox("Error occurred during GetLevelName" & vbNewLine & ex.Message, MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
            Return Nothing
        End Try
    End Function

    Private Sub chkboxBackup_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles chkboxBackup.CheckedChanged
        If chkboxBackup.Checked Then
            Me.txtBackupFrequency.Enabled = True
        Else
            Me.txtBackupFrequency.Enabled = False
        End If
    End Sub
End Class
